from dataclasses import dataclass
from typing import Any, Dict

from steamship.base import Client


@dataclass
class TrainPluginOutput:
    """
    This is the object produced by a completed training operation, stored as the `output` field of a `train` task.
    """

    # The tenant in which training is happening
    tenantId: str = None
    # The space in which training is happening
    spaceId: str = None
    #  The name of the model being trained
    modelName: str = None

    # The desired filename of the output model parameters.
    # This will be uploaded to the "Models" section of the Space's storage bucket
    modelFilename: str = None

    # A URL that has been pre-signed, permitting upload to that location
    modelUploadUrl: str = None

    # Arbitrary key-valued data to provide to the inference runner.
    # The training process will have the opportunity to amend this before writing it to the output
    inferenceParams: dict = None

    progress: dict = None
    resultData: dict = None
    error: str = None

    @staticmethod
    def from_dict(d: Any = None, client: Client = None) -> "TrainPluginOutput":
        if d is None:
            return None

        return TrainPluginOutput(
            tenantId=d.get("tenantId", None),
            spaceId=d.get("spaceId", None),
            modelName=d.get("modelName", None),
            modelFilename=d.get("modelFilename", None),
            modelUploadUrl=d.get("modelUploadUrl", None),
            inferenceParams=d.get("inferenceParams", None),
            progress=d.get("progress", None),
            resultData=d.get("resultData", None),
            error=d.get("error", None),
        )

    def to_dict(self) -> Dict:
        return dict(
            tenantId=self.tenantId,
            spaceId=self.spaceId,
            modelName=self.modelName,
            modelFilename=self.modelFilename,
            modelUploadUrl=self.modelUploadUrl,
            inferenceParams=self.inferenceParams,
            progress=self.progress,
            resultData=self.resultData,
            error=self.error,
        )



from dataclasses import dataclass
from typing import Any, List, Union, Optional

from steamship.base import Client, Request, Response
from steamship.base.request import IdentifierRequest
from steamship.data.tags.tag import Tag


@dataclass
class BlockQueryRequest(Request):
    tagFilterQuery: str


@dataclass
class Block:
    client: Client = None
    id: str = None
    fileId: str = None
    text: str = None
    tags: List[Tag] = None

    @dataclass
    class CreateRequest(Request):
        id: str = None
        fileId: str = None
        text: str = None
        tags: List[Tag.CreateRequest] = None
        upsert: bool = None

        # noinspection PyUnusedLocal
        @staticmethod
        def from_dict(d: Any, client: Client = None) -> "Block.CreateRequest":
            return Block.CreateRequest(
                id=d.get("id", None),
                fileId=d.get("fileId", None),
                text=d.get("text", None),
                tags=[Tag.CreateRequest.from_dict(tag) for tag in d.get("tags", [])],
                upsert=d.get("upsert", None),
            )

        def to_dict(self) -> dict:
            return dict(
                id=self.id,
                fileId=self.fileId,
                text=self.text,
                upsert=self.upsert,
                tags=[tag.to_dict() for tag in self.tags]
                if self.tags
                else [],  # TODO (enias): Deprecate
            )

    @dataclass
    class DeleteRequest(Request):
        id: str = None

    @dataclass
    class ListRequest(Request):
        fileId: str = None

    @dataclass
    class ListResponse(Request):
        blocks: List["Block"] = None

        @staticmethod
        def from_dict(d: Any, client: Client = None) -> "Optional[Block.ListResponse]":
            if d is None:
                return None
            return Block.ListResponse(
                blocks=[
                    Block.from_dict(x, client=client)
                    for x in (d.get("blocks", []) or [])
                ]
            )

    @staticmethod
    def from_dict(d: Any, client: Client = None) -> Union["Block", None]:
        if d is None:
            return None
        return Block(
            client=client,
            id=d.get("id", None),
            fileId=d.get("fileId", None),
            text=d.get("text", None),
            tags=list(
                map(lambda tag: Tag.from_dict(tag, client=client), d.get("tags", []))
            ),
        )

    def to_dict(self) -> dict:
        tags = None
        if self.tags is not None:
            tags = [tag.to_dict() for tag in self.tags]

        return dict(id=self.id, fileId=self.fileId, text=self.text, tags=tags)

    @staticmethod
    def get(
        client: Client,
        id: str = None,
        space_id: str = None,
        space_handle: str = None,
        space: Any = None,
    ) -> Response["Block"]:
        return client.post(
            "block/get",
            IdentifierRequest(id=id),
            expect=Block,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        )

    @staticmethod
    def create(
        client: Client,
        file_id: str = None,
        text: str = None,
        tags: List[Tag.CreateRequest] = None,
        upsert: bool = None,
        space_id: str = None,
        space_handle: str = None,
    ) -> Response["Block"]:
        req = Block.CreateRequest(fileId=file_id, text=text, tags=tags, upsert=upsert)
        return client.post(
            "block/create",
            req,
            expect=Block,
            space_id=space_id,
            space_handle=space_handle,
        )

    @staticmethod
    def list_public(
        client: Client,
        file_id: str = None,
        space_id: str = None,
        space_handle: str = None,
    ) -> Response["Block.ListResponse"]:
        return client.post(
            "block/list",
            Block.ListRequest(fileId=file_id),
            expect=Block.ListResponse,
            space_id=space_id,
            space_handle=space_handle,
        )

    def delete(self) -> Response["Block"]:
        return self.client.post(
            "block/delete",
            Block.DeleteRequest(id=self.id),
            expect=Tag,
        )

    @staticmethod
    def query(
        client: Client,
        tag_filter_query: str,
        space_id: str = None,
        space_handle: str = None,
        space: Any = None,
    ) -> Response["BlockQueryResponse"]:
        # TODO: Is this a static method?
        req = BlockQueryRequest(tagFilterQuery=tag_filter_query)
        res = client.post(
            "block/query",
            payload=req,
            expect=BlockQueryResponse,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        )
        return res


@dataclass
class BlockQueryResponse:
    blocks: List[Block]

    @staticmethod
    def from_dict(d: Any, client: Client = None) -> "BlockQueryResponse":
        return BlockQueryResponse(
            blocks=[
                Block.from_dict(block, client=client) for block in d.get("blocks", [])
            ]
        )



from steamship.app import Invocation, Request
from tests.demo_apps.apps.hello_world import HelloWorld, handler

NAME = "Ted"
RES_EMPTY = "Hello, Person"
RES_NAME = f"Hello, {NAME}"


def test_invoke_app_in_python():
    app = HelloWorld()

    assert app.greet().data == RES_EMPTY
    assert app.greet(NAME).data == RES_NAME


def test_invoke_app_with_request():
    app = HelloWorld()

    req = Request(invocation=Invocation(httpVerb="POST", appPath="greet"))
    res = app(req)
    assert res.data == RES_EMPTY

    req = Request(
        invocation=Invocation(
            httpVerb="POST", appPath="greet", arguments=dict(name=NAME)
        )
    )
    res = app(req)
    assert res.data == RES_NAME


def test_invoke_app_with_handler():
    event = dict(invocation=dict(httpVerb="POST", appPath="greet"))
    res = handler(event)
    assert res["data"] == RES_EMPTY

    event = dict(
        invocation=dict(httpVerb="POST", appPath="greet", arguments=dict(name=NAME))
    )
    res = handler(event)
    assert res["data"] == RES_NAME



import json
import logging
from dataclasses import asdict
from typing import Any, Dict, TypeVar, Union

import requests  # type: ignore

from steamship.base.configuration import Configuration
from steamship.base.error import SteamshipError
from steamship.base.mime_types import MimeTypes
from steamship.base.request import Request
from steamship.base.response import Response, Task

__copyright__ = "Steamship"
__license__ = "MIT"

_logger = logging.getLogger(__name__)

T = TypeVar("T", bound=Response)


class Client:
    """Client base.py class.

    Separated primarily as a hack to prevent circular imports.
    """

    # A client is always scoped by its space. A null space resolves to the
    # default space on the
    config: Configuration = None

    # Interaction prototype.
    dQuery: bool = False

    def __init__(  # TODO (Enias): Do we need all the default parameters?
        self,
        api_key: str = None,
        api_base: str = None,
        app_base: str = None,
        space_id: str = None,
        space_handle: str = None,
        profile: str = None,
        config_file: str = None,
        config_dict: dict = None,
        d_query: bool = False,
    ):

        self.config = Configuration(
            api_key=api_key,
            api_base=api_base,
            app_base=app_base,
            space_id=space_id,
            space_handle=space_handle,
            profile=profile,
            config_file=config_file,
            config_dict=config_dict,
        )

        self.dQuery = d_query

    def _url(
        self,
        app_call: bool = False,
        app_owner: str = None,
        operation: str = None,
        config: Configuration = None,
    ):  # TODO (enias): Simplify
        if not app_call:
            # Regular API call
            base = None
            if self.config and self.config.api_base:
                base = self.config.api_base
            if config and config.api_base:
                base = config.api_base
            if base is None:
                return SteamshipError(
                    code="EndpointMissing",
                    message="Can not invoke endpoint without the Client variable set.",
                    suggestion="This should automatically have a good default setting. "
                    "Reach out to our Steamship support.",
                )
        else:
            # Do the app version
            if app_owner is None:
                return SteamshipError(
                    code="UserMissing",
                    message="Can not invoke an app endpoint without the app owner's user handle.",
                    suggestion="Provide the appOwner option, or initialize your app with a lookup.",
                )

            base = None
            if self.config and self.config.app_base:
                base = self.config.app_base
            if config and config.app_base:
                base = config.app_base
            if base is None:
                return SteamshipError(
                    code="EndpointMissing",
                    message="Can not invoke an app endpoint without the App Base variable set.",
                    suggestion="This should automatically have a good default setting. "
                    "Reach out to our Steamship support.",
                )
            if (
                "localhost" not in base
                and "127.0.0.1" not in base
                and "0:0:0:0" not in base
            ):
                # We want to prepend the user handle
                parts = base.split("//")
                base = f"{parts[0]}//{app_owner}.{'//'.join(parts[1:])}"

        if base[len(base) - 1] == "/":
            base = base[:-1]
        if operation[0] == "/":
            operation = operation[1:]

        return f"{base}/{operation}"

    def _headers(
        self,
        space_id: str = None,
        space_handle: str = None,
        app_call: bool = False,
        app_owner: str = None,
        app_id: str = None,
        app_instance_id: str = None,
    ):
        headers = {"Authorization": f"Bearer {self.config.api_key}"}

        sid = space_id or self.config.space_id
        shandle = space_handle or self.config.space_handle

        if sid:
            headers["X-Space-Id"] = sid
        elif shandle:
            headers["X-Space-Handle"] = shandle

        if app_call:
            if app_owner:
                headers["X-App-Owner-Handle"] = app_owner
            if app_id:
                headers["X-App-Id"] = app_id
            if app_instance_id:
                headers["X-App-Instance-Id"] = app_instance_id
        return headers

    @staticmethod
    def _data(verb: str, file: Any, payload: Union[Request, dict]):
        if payload is None:
            data = {}
        elif isinstance(payload, dict):
            data = payload
        else:
            # noinspection PyDataclass
            data = asdict(payload)

        if verb == "POST" and file is not None:
            # Note: requests seems to have a bug passing boolean (and maybe numeric?)
            # values in the midst of multipart form data. You need to manually convert
            # it to a string; otherwise it will pass as False or True (with the capital),
            # which is not standard notation outside of Python.
            for key in data:
                if data[key] is False:
                    data[key] = "false"
                elif data[key] is True:
                    data[key] = "true"

        return data

    @staticmethod
    def _response_data(resp, raw_response: bool = False):
        if resp is None:
            return None

        if raw_response:
            return resp.content

        if resp.headers:
            ct = None
            if "Content-Type" in resp.headers:
                ct = resp.headers["Content-Type"]
            if "content-type" in resp.headers:
                ct = resp.headers["content-type"]
            if ct is not None:
                ct = ct.split(";")[0]  # application/json; charset=utf-8
                if ct in [MimeTypes.TXT, MimeTypes.MKD, MimeTypes.HTML]:
                    return resp.text
                elif ct == MimeTypes.JSON:
                    return resp.json()
                else:
                    return resp.content

    @staticmethod
    def make_file_dict(data, file):
        # TODO (enias): Review
        result = {}
        for key, val in data.items():
            if val:
                if isinstance(val, Dict):
                    result[key] = (None, json.dumps(val), "application/json")
                else:
                    result[key] = (None, str(val))
        result["file"] = file
        return result

    def call(
        self,
        verb: str,
        operation: str,
        payload: Union[Request, dict] = None,
        file: Any = None,
        expect: Any = None,
        asynchronous: bool = False,
        debug: bool = False,
        space_id: str = None,
        space_handle: str = None,
        space: Any = None,
        id_query: bool = None,
        raw_response: bool = False,
        app_call: bool = False,
        app_owner: str = None,
        app_id: str = None,
        app_instance_id: str = None,  # TODO (Enias): Where is the app_version_id ?
    ) -> Union[Any, Response[T]]:
        # TODO (Enias): Make shorter
        # TODO (Enias): Review naming convention
        """Post to the Steamship API.

        All responses have the format:
           {
             data: <actual response>,
             error?: {
               reason: message
             }
           }

        For the Python client we return the contents of the `data`
        field if present, and we raise an exception if the `error`
        field is filled in.
        """
        if self.config.api_key is None:
            raise Exception("Please set your Steamship API key.")

        if space_id is None and space is not None and hasattr(space, "id"):
            space_id = getattr(space, "id")

        if (
            space_id is None
            and space_handle is None
            and space is not None
            and hasattr(space, "handle")
        ):
            # Backup, if the spaceId transfer was None
            space_handle = getattr(space, "handle")

        url = self._url(
            app_call=app_call,
            app_owner=app_owner,
            operation=operation,
        )

        headers = self._headers(
            space_id=space_id,
            space_handle=space_handle,
            app_call=app_call,
            app_owner=app_owner,
            app_id=app_id,
            app_instance_id=app_instance_id,
        )

        data = self._data(verb=verb, file=file, payload=payload)

        if verb == "POST":
            if file is not None:
                files = self.make_file_dict(data, file)
                resp = requests.post(url, files=files, headers=headers)
            else:
                resp = requests.post(url, json=data, headers=headers)
        elif verb == "GET":
            resp = requests.get(url, params=data, headers=headers)
        else:
            raise Exception(f"Unsupported verb: {verb}")

        if debug is True:
            print("Response", resp)

        response_data = self._response_data(resp, raw_response=raw_response)

        if debug is True:
            print("Response JSON", response_data)

        task = None
        error = None
        obj = response_data

        if type(response_data) == dict:
            if "status" in response_data:
                task = Task.from_dict(response_data["status"], client=self)
                # if task_resp is not None and task_resp.taskId is not None:
                #     task = Task(client=self)
                #     task.update(task_resp)
                if "state" in response_data["status"]:
                    if response_data["status"]["state"] == "failed":
                        error = SteamshipError.from_dict(response_data["status"])

            if "data" in response_data:
                if expect is not None and hasattr(expect, "from_dict"):
                    obj = expect.from_dict(response_data["data"], client=self)
                else:
                    obj = response_data["data"]

            if "reason" in response_data:
                # This is a legacy error reporting field. We should work toward being comfortable
                # removing this handler.
                error = SteamshipError(message=response_data["reason"])

        ret_obj = None
        if error is None:
            ret_obj = obj

        ret = Response[T](
            expect=expect, task=task, data=ret_obj, error=error, client=self
        )

        if ret.task is None and ret.data is None and ret.error is None:
            raise Exception("No data, task status, or error found in response")

        if self.dQuery is True and asynchronous:
            # This is an experimental UI for jQuery-style chaining.
            ret.wait()
            # In dQuery mode we throw an error to stop the chain
            if ret.error is not None:
                raise ret.error

        if self.dQuery is True and id_query is not None:
            if ret.error is not None:
                raise ret.error
            return id_query

        return ret

    def post(
        self,
        operation: str,
        payload: Union[Request, dict] = None,
        file: Any = None,
        expect: Any = None,
        asynchronous: bool = False,
        debug: bool = False,
        space_id: str = None,
        space_handle: str = None,
        space: Any = None,
        id_query: bool = None,
        raw_response: bool = False,
        app_call: bool = False,
        app_owner: str = None,
        app_id: str = None,
        app_instance_id: str = None,
    ) -> Union[Any, Response[T]]:
        return self.call(
            verb="POST",
            operation=operation,
            payload=payload,
            file=file,
            expect=expect,
            asynchronous=asynchronous,
            debug=debug,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
            id_query=id_query,
            raw_response=raw_response,
            app_call=app_call,
            app_owner=app_owner,
            app_id=app_id,
            app_instance_id=app_instance_id,
        )

    def get(
        self,
        operation: str,
        payload: Union[Request, dict] = None,
        file: Any = None,
        expect: Any = None,
        asynchronous: bool = False,
        debug: bool = False,
        space_id: str = None,
        space_handle: str = None,
        space: Any = None,
        id_query: bool = None,
        raw_response: bool = False,
        app_call: bool = False,
        app_owner: str = None,
        app_id: str = None,
        app_instance_id: str = None,
    ) -> Union[Any, Response[T]]:
        return self.call(
            verb="GET",
            operation=operation,
            payload=payload,
            file=file,
            expect=expect,
            asynchronous=asynchronous,
            debug=debug,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
            id_query=id_query,
            raw_response=raw_response,
            app_call=app_call,
            app_owner=app_owner,
            app_id=app_id,
            app_instance_id=app_instance_id,
        )



import json

from steamship import MimeTypes
from steamship.data.block import Block
from steamship.data.file import File
from steamship.data.tags.tag import Tag

__copyright__ = "Steamship"
__license__ = "MIT"

from tests.utils.client import get_steamship_client


def test_file_upload():
    steamship = get_steamship_client()
    a = steamship.upload(content="A", mime_type=MimeTypes.MKD).data
    assert a.id is not None
    assert a.mime_type == MimeTypes.MKD

    b = steamship.upload(content="B", mime_type=MimeTypes.TXT).data
    assert b.id is not None
    assert b.mime_type == MimeTypes.TXT
    assert a.id != b.id

    c = steamship.upload(content="B", mime_type=MimeTypes.MKD).data
    assert (
        c.mime_type == MimeTypes.MKD
    )  # The specified format gets precedence over filename

    d = steamship.upload(
        content="B",
    ).data
    assert d.mime_type == MimeTypes.TXT  # The filename is used in a pinch.

    a.delete()
    b.delete()
    c.delete()
    d.delete()


def test_file_scrape():
    steamship = get_steamship_client()

    a = steamship.scrape(url="https://edwardbenson.com/2020/10/gpt3-travel-agent").data
    assert a.id is not None
    assert a.mime_type == MimeTypes.HTML

    b = steamship.scrape(
        url="https://edwardbenson.com/2018/09/case-of-the-murderous-ai"
    ).data
    assert b.id is not None
    assert a.id != b.id
    assert b.mime_type == MimeTypes.HTML

    a.delete()
    b.delete()


def test_file_import_response_dict():
    resp = File.CreateResponse(bytes=b"some bytes", mime_type=MimeTypes.BINARY)
    to_dict = resp.to_dict()
    from_dict = File.CreateResponse.from_dict(to_dict)
    assert resp.data == from_dict.data
    assert resp.mimeType == from_dict.mimeType


def test_file_import_response_bytes_serialization():
    file_resp = File.CreateResponse(bytes=b"some bytes", mime_type=MimeTypes.BINARY)
    to_dict = file_resp.to_dict()
    as_json_string = json.dumps(to_dict)
    as_dict_again = json.loads(as_json_string)
    assert as_dict_again == to_dict


def test_file_upload_with_blocks():
    client = get_steamship_client()
    a = File.create(
        client=client,
        blocks=[
            Block.CreateRequest(text="A", tags=[Tag.CreateRequest(name="BlockTag")]),
            Block.CreateRequest(text="B"),
        ],
        tags=[Tag.CreateRequest(name="FileTag")],
    ).data
    assert a.id is not None

    blocks = Block.list_public(client, file_id=a.id)

    def check_blocks(block_list):
        assert len(block_list) == 2
        assert block_list[0].tags is not None
        assert len(block_list[0].tags) == 1
        assert block_list[0].tags[0].name == "BlockTag"
        assert block_list[0].text == "A"

    assert blocks.data.blocks is not None
    check_blocks(blocks.data.blocks)

    # Let's get the file fresh
    aa = File.get(client, id=a.id).data
    check_blocks(aa.blocks)
    assert aa.tags is not None
    assert len(aa.tags) == 1
    assert aa.tags[0].name == "FileTag"

    a.delete()


def test_query():
    client = get_steamship_client()
    a = File.create(
        client=client,
        blocks=[
            Block.CreateRequest(text="A", tags=[Tag.CreateRequest(name="BlockTag")]),
            Block.CreateRequest(text="B"),
        ],
    ).data
    assert a.id is not None
    b = File.create(
        client=client,
        blocks=[Block.CreateRequest(text="A"), Block.CreateRequest(text="B")],
        tags=[Tag.CreateRequest(name="FileTag")],
    ).data
    assert b.id is not None

    files = File.query(
        client=client, tag_filter_query='blocktag and name "BlockTag"'
    ).data.files
    assert len(files) == 1
    assert files[0].id == a.id

    files = File.query(
        client=client, tag_filter_query='filetag and name "FileTag"'
    ).data.files
    assert len(files) == 1
    assert files[0].id == b.id

    a.delete()
    b.delete()



import logging
from dataclasses import dataclass
from typing import Any, Union


@dataclass
class SteamshipError(Exception):
    message: str = None
    internalMessage: str = None
    suggestion: str = None
    code: str = None
    error: str = None

    def __init__(
        self,
        message: str = "Undefined remote error",
        internal_message: str = None,
        suggestion: str = None,
        code: str = None,
        error: Union[Exception, str] = None,
    ):
        self.message = message
        self.suggestion = suggestion
        self.internalMessage = internal_message
        self.statusCode = code
        if error is not None:
            self.error = str(error)

        parts = []
        if code is not None:
            parts.append(f"[{code}]")
        if message is not None:
            parts.append(message)
        if internal_message is not None:
            parts.append(f"Internal Message: {internal_message}")
        if suggestion is not None:
            parts.append(f"Suggestion: {suggestion}")

        super().__init__("\n".join(parts))

    def log(self):
        logging.error(
            f"[{self.code}] {self.message}. [Internal: {self.internalMessage}] [Suggestion: { self.suggestion}]"
        )
        if self.error:
            logging.error(self.error)

    def to_dict(self) -> dict:
        return dict(
            message=self.message,
            internalMessage=self.internalMessage,
            suggestion=self.suggestion,
            code=self.code,
            error=self.error,
        )

    @staticmethod
    def from_dict(d: Any) -> "SteamshipError":
        """Last resort if subclass doesn't override: pass through."""
        return SteamshipError(
            message=d.get("statusMessage", d.get("message", None)),
            internal_message=d.get("internalMessage", None),
            suggestion=d.get("statusSuggestion", d.get("suggestion", None)),
            code=d.get("statusCode", d.get("code", None)),
            error=d.get("error", d.get("error", None)),
        )



from dataclasses import dataclass
from typing import Any

from steamship.base import Client, Request, Response
from steamship.base.request import GetRequest, IdentifierRequest


class SignedUrl:
    class Bucket:
        exports = "exports"
        imports = "imports"
        userData = "userData"
        pluginData = "pluginData"
        appData = "appData"
        models = "models"

    class Operation:
        read = "Read"
        write = "Write"

    @dataclass
    class Request(Request):
        bucket: str = None
        filepath: str = None
        operation: str = None
        expiresInMinutes: int = None

    @dataclass
    class Response(Response):
        bucket: str = None
        filepath: str = None
        operation: str = None
        expiresInMinutes: int = None
        signedUrl: str = None


@dataclass
class Space:
    client: Client = None
    id: str = None
    handle: str = None

    @dataclass
    class CreateRequest(Request):
        id: str = None
        handle: str = None
        upsert: bool = None
        externalId: str = None
        externalType: str = None
        metadata: str = None

    @dataclass
    class ListRequest(Request):
        pass

    def delete(self) -> "Response[Space]":
        return self.client.post(
            "space/delete", IdentifierRequest(id=self.id), expect=Space
        )

    @staticmethod
    def from_dict(d: Any, client: Client) -> "Space":
        if "space" in d:
            d = d["space"]

        return Space(client=client, id=d.get("id", None), handle=d.get("handle", None))

    @staticmethod
    def get(
        client: Client,
        id_: str = None,
        handle: str = None,
        upsert: bool = None,
        space_id: str = None,
        space_handle: str = None,
        space: "Space" = None,
    ) -> "Response[Space]":
        req = GetRequest(id=id_, handle=handle, upsert=upsert)
        return client.post(
            "space/get",
            req,
            expect=Space,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        )

    @staticmethod
    def create(
        client: Client,
        handle: str,
        external_id: str = None,
        external_type: str = None,
        metadata: Any = None,
        upsert: bool = True,
    ) -> "Response[Space]":
        req = Space.CreateRequest(
            handle=handle,
            upsert=upsert,
            externalId=external_id,
            externalType=external_type,
            metadata=metadata,
        )
        return client.post("space/create", req, expect=Space)

    def create_signed_url(
        self, request: SignedUrl.Request
    ) -> Response[SignedUrl.Response]:
        return self.client.post(
            "space/createSignedUrl", payload=request, expect=SignedUrl.Response
        )



from dataclasses import dataclass
from typing import Any, Dict, Optional

from steamship.base import Client


@dataclass
class TrainPluginInput:
    """
    This is the object passed as input to a training operation, stored as the `input` field of a `train` task.
    """

    # The tenant in which training is happening
    tenantId: str = None
    # The space in which training is happening
    spaceId: str = None

    # The plugin instance being trained (this is the handle)
    pluginInstance: str = None
    # The plugin instance being trained (this is the UUID)
    pluginInstanceId: str = None

    #  The name of the model to train; e.g. "DistilBERT"
    modelName: str = None

    # The desired filename of the output model parameters.
    # This will be uploaded to the "Models" section of the Space's storage bucket
    modelFilename: str = None

    # A URL that has been pre-signed, permitting upload to that location
    modelUploadUrl: str = None

    # How may epochs of training to perform, if relevant and supported
    trainingEpochs: int = None
    # How much data to hold out for testing & reporting, if relevant and supported.
    testingHoldoutPercent: float = None
    # An optional seed for the train-test split
    testSplitSeed: int = None

    # Arbitrary key-valued data to provide to the particular `modelName` trainer.
    trainingParams: dict = None
    # Arbitrary key-valued data to provide to the inference runner in the TrainPluginOutput object.
    # The training process will have the opportunity to amend this before writing it to the output
    inferenceParams: dict = None

    # A pre-signed URL at which the training data can be found
    trainingDataUrl: str = None

    @staticmethod
    def from_dict(d: Any = None, client: Client = None) -> "Optional[TrainPluginInput]":
        if d is None:
            return None

        return TrainPluginInput(
            tenantId=d.get("tenantId", None),
            spaceId=d.get("spaceId", None),
            pluginInstance=d.get("pluginInstance", None),
            pluginInstanceId=d.get("pluginInstanceId", None),
            modelName=d.get("modelName", None),
            modelFilename=d.get("modelFilename", None),
            modelUploadUrl=d.get("modelUploadUrl", None),
            trainingEpochs=d.get("trainingEpochs", None),
            testingHoldoutPercent=d.get("testingHoldoutPercent", None),
            testSplitSeed=d.get("testSplitSeed", None),
            trainingParams=d.get("trainingParams", None),
            inferenceParams=d.get("inferenceParams", None),
            trainingDataUrl=d.get("trainingDataUrl", None),
        )

    def to_dict(self) -> Dict:
        return dict(
            tenantId=self.tenantId,
            spaceId=self.spaceId,
            pluginInstance=self.pluginInstance,
            pluginInstanceId=self.pluginInstanceId,
            modelName=self.modelName,
            modelFilename=self.modelFilename,
            modelUploadUrl=self.modelUploadUrl,
            trainingEpochs=self.trainingEpochs,
            testingHoldoutPercent=self.testingHoldoutPercent,
            testSplitSeed=self.testSplitSeed,
            trainingParams=self.trainingParams,
            inferenceParams=self.inferenceParams,
            trainingDataUrl=self.trainingDataUrl,
        )



from steamship import MimeTypes, PluginInstance

__copyright__ = "Steamship"
__license__ = "MIT"

# TODO: It should fail if the docs field is empty.
# TODO: It should fail if the file hasn't been converted.
from tests.utils.client import get_steamship_client
from tests.utils.random import random_index

_TEST_EMBEDDER = "test-embedder"

T = "A nice poem"
P1_1 = "Roses are red."
P1_2 = "Violets are blue."
P2_1 = "Sugar is sweet."
P2_2 = "I love you."
T2 = "A flavorful story"
P3_1 = "Cake is made of flour."
P3_2 = "Cake tastes good with milk."
P4_1 = "Cake comes in chocolate and vanilla flavors."
P4_2 = "Cake can be cut into mAny pieces and shared."


def test_file_parse():
    steamship = get_steamship_client()
    content1 = "# {}\n\n{} {}\n\n{} {}".format(T, P1_1, P1_2, P2_1, P2_2)
    content2 = "# {}\n\n{} {}\n\n{} {}".format(T2, P3_1, P3_2, P4_1, P4_2)
    content = "{}\n\n{}".format(content1, content2)

    a = steamship.upload(content=content, mime_type=MimeTypes.MKD).data
    assert a.id is not None
    assert a.mime_type == MimeTypes.MKD

    blockify_resp = a.blockify(plugin_instance="markdown-blockifier-default-1.0")
    assert blockify_resp.error is None
    blockify_resp.wait()

    # Now we parse
    parser = PluginInstance.create(steamship, plugin_handle="test-tagger").data
    parse_resp = a.tag(plugin_instance=parser.handle)
    assert parse_resp.error is None
    parse_resp.wait()

    # Now the sentences should be parsed!
    q2 = a.refresh().data
    assert len(q2.blocks) == 6

    # Now we add the file to the index
    plugin_instance = PluginInstance.create(
        steamship, plugin_handle=_TEST_EMBEDDER
    ).data
    with random_index(steamship, plugin_instance=plugin_instance.handle) as index:
        index.insert_file(a.id, reindex=False)
        embed_resp = index.embed()
        assert embed_resp.error is None
        embed_resp.wait()

        res = index.search("What color are roses?").data
        assert len(res.items) == 1
        # Because the simdex now indexes entire blocks and not sentences, the result of this is the whole block text
        assert res.items[0].value.value == " ".join([P1_1, P1_2])

    a.delete()


def test_file_index():
    steamship = get_steamship_client()
    t = "A nice poem"
    p1_1 = "Roses are red."
    p1_2 = "Violets are blue."
    p2_1 = "Sugar is sweet."
    p2_2 = "I love you."
    t2 = "A flavorful story"
    p3_1 = "Cake is made of flour."
    p3_2 = "Cake tastes good with milk."
    p4_1 = "Cake comes in chocolate and vanilla flavors."
    p4_2 = "Cake can be cut into mAny pieces and shared."

    content1 = f"# {t}\n\n{p1_1} {p1_2}\n\n{p2_1} {p2_2}"
    content2 = f"# {t2}\n\n{p3_1} {p3_2}\n\n{p4_1} {p4_2}"
    content = f"{content1}\n\n{content2}"

    a = steamship.upload(content=content, mime_type=MimeTypes.MKD).data
    assert a.id is not None
    assert a.mime_type == MimeTypes.MKD

    blockify_resp = a.blockify(plugin_instance="markdown-blockifier-default-1.0")
    assert blockify_resp.error is None
    blockify_resp.wait()

    # Now we parse
    parser = PluginInstance.create(steamship, plugin_handle="test-tagger").data
    parse_resp = a.tag(plugin_instance=parser.handle)
    assert parse_resp.error is None
    parse_resp.wait()

    # Now the sentences should be parsed!
    q2 = a.refresh().data
    assert len(q2.blocks) == 6

    # Now we add the file to the index via the shortcut.
    embedder = PluginInstance.create(steamship, plugin_handle="test-embedder").data
    index = a.index(plugin_instance=embedder.handle)

    res = index.search("What color are roses?").data
    assert len(res.items) == 1
    # Because the simdex now indexes entire blocks and not sentences, the result of this is the whole block text
    assert res.items[0].value.value == " ".join([p1_1, p1_2])

    res = index.search("What flavors does cake come in?").data
    assert len(res.items) == 1
    # Because the simdex now indexes entire blocks and not sentences, the result of this is the whole block text
    assert res.items[0].value.value == " ".join([p4_1, p4_2])

    index.delete()
    a.delete()


def test_file_embed_lookup():
    steamship = get_steamship_client()

    content_a = "Ted likes to run."
    content_b = "Grace likes to bike."

    a = steamship.upload(content=content_a, mime_type=MimeTypes.MKD).data

    blockify_res = a.blockify(plugin_instance="markdown-blockifier-default-1.0")
    assert blockify_res.error is None
    blockify_res.wait()

    parser = PluginInstance.create(steamship, plugin_handle="test-tagger").data
    parse_res = a.tag(plugin_instance=parser.handle)
    assert parse_res.error is None
    parse_res.wait()

    b = steamship.upload(content=content_b, mime_type=MimeTypes.MKD).data
    blockify_res = b.blockify(plugin_instance="markdown-blockifier-default-1.0")
    assert blockify_res.error is None
    blockify_res.wait()

    parser = PluginInstance.create(steamship, plugin_handle="test-tagger").data
    parse_res = b.tag(plugin_instance=parser.handle)
    assert parse_res.error is None
    parse_res.wait()

    embedder = PluginInstance.create(steamship, plugin_handle="test-embedder").data
    # Now we add the file to the index
    with random_index(steamship, embedder.handle) as index:
        index.insert_file(a.id, block_type="sentence", reindex=True)
        index.insert_file(b.id, block_type="sentence", reindex=True)

        res = index.search("What does Ted like to do?").data
        assert len(res.items) == 1
        assert res.items[0].value.value == content_a

        res = index.search("What does Grace like to do?").data
        assert len(res.items) == 1
        assert res.items[0].value.value == content_b

        # Now we list the items
        itemsa = index.list_items(file_id=a.id).data
        for item in itemsa.items:
            print("File {} - Value {}".format(item.fileId, item.value))
        assert len(itemsa.items) == 1
        assert len(itemsa.items[0].embedding) > 0
        assert itemsa.items[0].value == content_a

        itemsb = index.list_items(file_id=b.id).data
        assert len(itemsb.items) == 1
        assert len(itemsb.items[0].embedding) > 0
        assert len(itemsb.items[0].embedding) == len(itemsa.items[0].embedding)
        assert itemsb.items[0].value == content_b



import logging
from typing import Any, List

from steamship import Block
from steamship.base import Client, Response
from steamship.client.operations.tagger import TagRequest
from steamship.client.tasks import Tasks
from steamship.data import File
from steamship.data.embeddings import (
    EmbedAndSearchRequest,
    EmbeddingIndex,
    QueryResults,
)
from steamship.data.space import Space

__copyright__ = "Steamship"
__license__ = "MIT"

from steamship.extension.file import TagResponse

_logger = logging.getLogger(__name__)


class Steamship(Client):
    """Steamship Python Client."""

    def __init__(
        self,
        api_key: str = None,
        api_base: str = None,
        app_base: str = None,
        space_id: str = None,
        space_handle: str = None,
        profile: str = None,
        config_file: str = None,
        config_dict: dict = None,
        d_query: bool = False,
    ):  # TODO (Enias): What is d_query?
        # TODO (Enias): Do we need config file and config dict?
        super().__init__(
            api_key=api_key,
            api_base=api_base,
            app_base=app_base,
            space_id=space_id,
            space_handle=space_handle,
            profile=profile,
            config_file=config_file,
            config_dict=config_dict,
            d_query=d_query,
        )
        """
        The base.py class will properly detect and set the defaults. They should be None here.

        dQuery is a Beta option that will return chainable responses, like:
          file.upload()
              .blockify()
              .parse()
              .embed()
              .query()

        It offers no new functionality -- in fact at the moment it's slightly less in that you
        are given the syntactically convenient response object for chaining rather than the actual
        response object of the invocation.
        """
        self.tasks = Tasks(self)

    def create_index(
        self,
        handle: str = None,
        plugin_instance: str = None,
        upsert: bool = True,
        external_id: str = None,
        external_type: str = None,
        metadata: Any = None,
        space_id: str = None,
        space_handle: str = None,
        space: Space = None,
    ) -> Response[EmbeddingIndex]:
        return EmbeddingIndex.create(
            client=self,
            handle=handle,
            plugin_instance=plugin_instance,
            upsert=upsert,
            external_id=external_id,
            external_type=external_type,
            metadata=metadata,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        )

    def upload(
        self,
        filename: str = None,
        content: str = None,
        mime_type: str = None,
        space_id: str = None,
        space_handle: str = None,
        space: Space = None,
    ) -> Response[File]:
        return File.create(
            self,
            filename=filename,
            content=content,
            mime_type=mime_type,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        )

    def scrape(
        self,
        url: str,
        space_id: str = None,
        space_handle: str = None,
        space: Space = None,
    ) -> Response[File]:
        return File.scrape(
            self, url, space_id=space_id, space_handle=space_handle, space=space
        )

    def embed_and_search(
        self,
        query: str,
        docs: List[str],
        plugin_instance: str,
        k: int = 1,
        space_id: str = None,
        space_handle: str = None,
        space: Space = None,
    ) -> Response[QueryResults]:
        req = EmbedAndSearchRequest(
            query=query, docs=docs, pluginInstance=plugin_instance, k=k
        )
        return self.post(
            "plugin/instance/embeddingSearch",
            req,
            expect=QueryResults,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        )

    def tag(
        self,
        doc: str,
        plugin_instance: str = None,
        space_id: str = None,
        space_handle: str = None,
        space: Space = None,
    ) -> Response[TagResponse]:
        req = TagRequest(
            type="inline",
            file=File.CreateRequest(blocks=[Block.CreateRequest(text=doc)]),
            pluginInstance=plugin_instance,
        )
        return self.post(
            "plugin/instance/tag",
            req,
            expect=TagResponse,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        )



from dataclasses import dataclass
from typing import Any

from steamship.app import Response
from steamship.base import Client


@dataclass
class User:
    client: Client = None
    id: str = None
    handle: str = None

    @staticmethod
    def from_dict(d: Any, client: Client = None) -> "User":
        if d["user"] is not None:
            d = d["user"]

        return User(
            client=client,
            id=d.get("id", None),
            handle=d.get("handle", None),
        )

    def to_dict(self) -> dict:
        return dict(id=self.id, handle=self.handle)

    @staticmethod
    def current(client: Client) -> "Response[User]":
        return client.get("account/current", expect=User)



from steamship import SteamshipError
from steamship.app import App, Response, create_handler, post
from steamship.plugin.inputs.block_and_tag_plugin_input import BlockAndTagPluginInput
from steamship.plugin.inputs.train_plugin_input import TrainPluginInput
from steamship.plugin.outputs.block_and_tag_plugin_output import BlockAndTagPluginOutput
from steamship.plugin.outputs.train_plugin_output import TrainPluginOutput
from steamship.plugin.outputs.training_parameter_plugin_output import (
    TrainingParameterPluginOutput,
)
from steamship.plugin.service import PluginRequest
from steamship.plugin.tagger import Tagger


class TestTrainableTaggerPlugin(Tagger, App):
    # For testing; mirrors TestConfigurableTagger in Swift

    RESPONSE = TrainingParameterPluginOutput(
        trainingEpochs=3,
        modelName="pytorch_text_classification",
        testingHoldoutPercent=0.3,
        trainingParams=dict(foo=1),
    )

    def run(
        self, request: PluginRequest[BlockAndTagPluginInput]
    ) -> Response[BlockAndTagPluginOutput]:
        raise SteamshipError(
            message="Inference on this tagger is performed by the Steamship Inference Cloud."
        )

    # noinspection PyUnusedLocal
    @post("getTrainingParameters")
    def get_training_parameters(
        self, **kwargs
    ) -> Response[TrainingParameterPluginOutput]:
        return Response(data=TestTrainableTaggerPlugin.RESPONSE)

    @post("train")
    def train(self, **kwargs) -> Response[TrainPluginInput]:
        train_plugin_input = TrainPluginInput.from_dict(kwargs)
        return Response(
            data=TrainPluginOutput(
                tenantId=train_plugin_input.tenantId,
                spaceId=train_plugin_input.spaceId,
                modelUploadUrl=train_plugin_input.modelUploadUrl,
                modelFilename=train_plugin_input.modelFilename,
            )
        )


handler = create_handler(TestTrainableTaggerPlugin)



from dataclasses import dataclass


class Request:
    pass


@dataclass
class GetRequest(Request):
    id: str = None
    handle: str = None
    upsert: bool = None


@dataclass
class IdentifierRequest(Request):
    id: str = None
    handle: str = None



from dataclasses import dataclass
from typing import Any, Dict

from steamship.base import Client, Request, Response


@dataclass
class CreateAppVersionRequest(Request):
    appId: str = None
    handle: str = None
    upsert: bool = None
    type: str = "file"
    configTemplate: Dict[str, Any] = None


@dataclass
class DeleteAppVersionRequest(Request):
    id: str


@dataclass
class ListPrivateAppVersionsRequest(Request):
    pass


@dataclass
class AppVersion:
    client: Client = None
    id: str = None
    appId: str = None
    handle: str = None
    configTemplate: Dict[str, Any] = None

    @staticmethod
    def from_dict(d: Any, client: Client = None) -> "AppVersion":
        if "appVersion" in d:
            d = d["appVersion"]

        return AppVersion(
            client=client,
            id=d.get("id", None),
            handle=d.get("handle", None),
            configTemplate=d.get("configTemplate", None),
        )

    @staticmethod
    def create(
        client: Client,
        app_id: str = None,
        handle: str = None,
        filename: str = None,
        filebytes: bytes = None,
        upsert: bool = None,
        config_template: Dict[str, Any] = None,
    ) -> "Response[AppVersion]":

        if filename is None and filebytes is None:
            raise Exception("Either filename or filebytes must be provided.")
        if filename is not None and filebytes is not None:
            raise Exception("Only either filename or filebytes should be provided.")

        if filename is not None:
            with open(filename, "rb") as f:
                filebytes = f.read()

        req = CreateAppVersionRequest(
            handle=handle, appId=app_id, upsert=upsert, configTemplate=config_template
        )

        return client.post(
            "app/version/create",
            payload=req,
            file=("app.zip", filebytes, "multipart/form-data"),
            expect=AppVersion,
        )

    def delete(self) -> "Response[AppVersion]":
        req = DeleteAppVersionRequest(id=self.id)
        return self.client.post("app/version/delete", payload=req, expect=AppVersion)



import dataclasses
import io
import logging
from dataclasses import dataclass
from typing import Any, Dict, Generic, TypeVar, Union

from steamship.base import SteamshipError
from steamship.base.binary_utils import flexi_create
from steamship.base.mime_types import ContentEncodings, MimeTypes
from steamship.base.tasks import Task, TaskState


@dataclass
class Http:
    status: int = None
    # If true, we're signaling to the Steamship Proxy that the `data` field of the SteamshipResponse object
    # has been wrapped in base64. In this situation, we can return the bytes within directly to the Proxy
    # caller without interpreting it.
    base64Wrapped: bool = None
    headers: Dict[str, str] = None


T = TypeVar("T")


@dataclass
class Response(Generic[T]):
    """Mirrors the Response object in the Steamship server."""

    data: T = None  # Data for successful or synchronous requests.
    status: Task = None  # Reporting for errors and async status
    http: Http = None  # Additional HTTP information for Steamship Proxy (headers, etc)

    def __init__(
        self,
        status: Task = None,
        error: SteamshipError = None,
        http: Http = None,
        data: Any = None,
        string: str = None,
        json: Any = None,
        bytes: Union[bytes, io.BytesIO] = None,
        mime_type=None,
    ):
        # Note:
        # This function has to be very defensively coded since Any errors thrown here will not be returned
        # to the end-user via our proxy (as this is the constructor for the response itself!)
        if http is not None:
            self.http = http
        else:
            self.http = Http(status=200, headers={})

        # Handle the core data
        try:
            data, mime_type, encoding = flexi_create(
                data=data, string=string, json=json, bytes=bytes, mime_type=mime_type
            )

            self.data = data

            if mime_type is None:
                mime_type = MimeTypes.BINARY

            if mime_type is not None:
                if self.http.headers is None:
                    self.http.headers = {}
                self.http.headers["Content-Type"] = mime_type

            if encoding == ContentEncodings.BASE64:
                self.http.base64Wrapped = True

        except Exception as ex:
            logging.error(f"Exception within Response.__init__. {ex}")
            if error is not None:
                if error.message:
                    error.message = f"{error.message}. Also found error - unable to serialize data to response. {ex}"
                else:
                    error.message = f"Unable to serialize data to response. {ex}"
            else:
                error = SteamshipError(
                    message=f"Unable to serialize data to response. {ex}"
                )
            logging.error(error)

        # Handle the task provided
        if status is None:
            self.status = Task()
        elif type(status) == Task:
            self.status = status
        else:
            self.status = Task()
            self.status.state = TaskState.failed
            self.status.status_message = (
                f"Status field of response should be of type Task. "
                f"Instead was of type {type(status)} and had value {status}."
            )

        if error:
            self.status.state = TaskState.failed
            self.status.status_message = error.message
            self.status.status_suggestion = error.suggestion
            self.status.status_code = error.code
            logging.error("steamship.app.response - Response created with error.")
            logging.error(error)
        else:
            if self.status.state is None:
                self.status.state = TaskState.succeeded

    @staticmethod
    def error(
        code: int,
        message: str = None,
        error: SteamshipError = None,
        exception: Exception = None,
    ) -> "Response[T]":
        error = error or SteamshipError(message=message)

        if error.message is None:
            error.message = message
        else:
            error.message = f"{error.message}. {message}"

        if exception is not None:
            if error.message is None:
                error.message = f"{exception}"
            else:
                error.message = f"{error.message}. {exception}"

        return Response(
            error=error or SteamshipError(message=message), http=Http(status=code)
        )

    @staticmethod
    def from_obj(obj: Any) -> "Response":
        if obj is None:
            return Response.error(500, "Handler provided no response.")

        obj_t = type(obj)

        if obj_t == Response:
            return obj
        elif obj_t == SteamshipError:
            return Response.error(500, error=obj)
        elif obj_t == Exception:
            return Response.error(500, error=SteamshipError(error=obj))
        elif obj_t == io.BytesIO:
            return Response(bytes=obj)
        elif obj_t == dict:
            return Response(json=obj)
        elif obj_t == str:
            return Response(string=obj)
        elif obj_t in [float, int, bool]:
            return Response(json=obj)

        if getattr(obj, "to_dict"):
            try:
                return Response(json=getattr(obj, "to_dict")())
            except Exception as e:
                logging.error(f"Failed calling to_dict on response object. {obj}\n {e}")

        if dataclasses.is_dataclass(obj):
            return Response(json=dataclasses.asdict(obj))

        return Response.error(500, message="Handler provided unknown response type.")

    def to_dict(self) -> Dict:
        return dataclasses.asdict(self)



from dataclasses import dataclass
from typing import Any, List

from steamship import File
from steamship.base import Client


@dataclass
class CorpusImportRequest:  # TODO (enias): Depricate
    # The Corpus Identifiers
    client: Client = None
    id: str = None
    handle: str = None
    type: str = "corpus"

    # Data for the plugin
    value: str = None
    data: str = None
    url: str = None
    pluginInstance: str = None
    fileImporterPluginInstance: str = None

    @staticmethod
    def from_dict(d: Any, client: Client = None) -> "CorpusImportRequest":
        # noinspection PyArgumentEqualDefault
        return CorpusImportRequest(
            client=client,
            id=d.get("id", None),
            handle=d.get("handle", None),
            type="corpus",
            value=d.get("value", None),
            data=d.get("data", None),
            url=d.get("url", None),
            pluginInstance=d.get("pluginInstance", None),
            fileImporterPluginInstance=d.get("fileImporterPluginInstance", None),
        )

    def to_dict(self) -> dict:
        return dict(
            id=self.id,
            handle=self.handle,
            type=self.type,
            value=self.value,
            data=self.data,
            url=self.url,
            pluginInstance=self.pluginInstance,
            fileImporterPluginInstance=self.fileImporterPluginInstance,
        )


@dataclass
class CorpusImportResponse:
    client: Client = None
    fileImportRequests: List[File.CreateRequest] = None

    def __init__(
        self,
        client: Client = None,
        file_import_requests: List[File.CreateRequest] = None,
    ):
        self.client = client
        self.fileImportRequests = file_import_requests

    @staticmethod
    def from_dict(d: Any, client: Client = None) -> "CorpusImportResponse":
        return CorpusImportResponse(
            client=client,
            file_import_requests=[
                File.CreateRequest.from_dict(req)
                for req in d.get("fileImportRequests", [])
            ],
        )

    def to_dict(self) -> dict:
        return dict(fileImportRequests=self.fileImportRequests)



from dataclasses import dataclass
from typing import Any, Dict

from steamship.base.configuration import Configuration


def event_to_config(event: dict) -> Configuration:
    if event is None:
        raise Exception("Null event provided")
    if "invocationContext" not in event:
        raise Exception("invocationContext not in event")

    return Configuration.from_dict(event["invocationContext"])


class Verb:
    GET = "GET"
    POST = "POST"

    @staticmethod
    def safely_from_str(s: str) -> str:
        ss = s.strip().upper()
        if ss == Verb.GET:
            return Verb.GET
        elif ss == Verb.POST:
            return Verb.POST
        return s


@dataclass
class Invocation:
    httpVerb: str = None
    appPath: str = None  # e.g. /hello/there
    arguments: Dict[str, Any] = None
    config: Dict[str, Any] = None

    @staticmethod
    def from_dict(d: dict) -> "Invocation":
        return Invocation(
            httpVerb=d.get("httpVerb", None),
            appPath=d.get("appPath", None),
            arguments=d.get("arguments", None),
            config=d.get("config", None),
        )


@dataclass
class Request:
    """An request of a method on an app instance.

    This is the payload sent from the public-facing App Proxy to the
    private-facing app microservice.
    """

    # TODO (enias): Is this a replacement for the Request in base?

    clientConfig: Configuration = None
    invocation: Invocation = None

    @staticmethod
    def from_dict(d: dict) -> "Request":
        invocation = Invocation.from_dict(d.get("invocation", dict()))
        client_config = Configuration.from_dict(
            d.get("clientConfig", dict())
        )  # TODO (enias): Review config dict
        return Request(clientConfig=client_config, invocation=invocation)



class TextTag:
    """This class provides a set of SpaCy-style constants to use with Tag(kind="token")."""

    textWithWs = "textWithWs"
    whitespace = "whitespace"
    head = "head"
    leftEdge = "leftEdge"
    rightEdge = "rightEdge"
    entType = "entType"
    entIob = "entIob"
    lemma = "lemma"
    normalized = "normalized"
    shape = "shape"
    prefix = "prefix"
    suffix = "suffix"
    isAlpha = "isAlpha"
    isAscii = "isAscii"
    isDigit = "isDigit"
    isTitle = "isTitle"
    isPunct = "isPunct"
    isLeftPunct = "isLeftPunct"
    isRightPunct = "isRightPunct"
    isSpace = "isSpace"
    isBracket = "isBracket"
    isQuote = "isQuote"
    isCurrency = "isCurrency"
    likeUrl = "likeUrl"
    likeNum = "likeNum"
    likeEmail = "likeEmail"
    isOov = "isOov"
    isStop = "isStop"
    lang = "lang"
    embedding = "embedding"



from typing import Any

from steamship import Tag
from steamship.base import Client, Response
from steamship.client.operations.blockifier import BlockifyRequest
from steamship.client.operations.tagger import TagRequest, TagResponse
from steamship.data.embeddings import EmbeddedItem, EmbeddingIndex
from steamship.data.file import File, FileUploadType
from steamship.data.plugin import PluginTargetType
from steamship.plugin.outputs.block_and_tag_plugin_output import BlockAndTagPluginOutput


def upload(
    client: Client,
    filename: str = None,
    content: str = None,
    mime_type: str = None,
    corpus_id: str = None,
    space_id: str = None,
    space_handle: str = None,
    space: Any = None,
) -> "Response[File]":
    if filename is None and content is None:
        raise Exception(
            "Either filename or content must be provided."
        )  # TODO (Enias): Review

    if filename is not None:
        with open(filename, "rb") as f:
            content = f.read()

    req = File.CreateRequest(
        type=FileUploadType.file, corpusId=corpus_id, mimeType=mime_type
    )

    return client.post(
        "file/create",
        payload=req,
        file=(content, "multipart/form-data"),
        expect=File,
        space_id=space_id,
        space_handle=space_handle,
        space=space,
    )


File.upload = upload


def blockify(self, plugin_instance: str = None):
    req = BlockifyRequest(type="file", id=self.id, pluginInstance=plugin_instance)

    return self.client.post(
        "plugin/instance/blockify",
        payload=req,
        expect=BlockAndTagPluginOutput,
        asynchronous=True,
        id_query=self,
    )


File.blockify = blockify  # TODO (enias): Review


def tag(
    self,
    plugin_instance: str = None,
    space_id: str = None,
    space_handle: str = None,
    space: Any = None,
) -> Response[Tag]:
    req = TagRequest(
        type=PluginTargetType.file, id=self.id, pluginInstance=plugin_instance
    )

    return self.client.post(
        "plugin/instance/tag",
        payload=req,
        expect=TagResponse,
        asynchronous=True,
        id_query=self,
        space_id=space_id,
        space_handle=space_handle,
        space=space,
    )


File.tag = tag


def index(
    self,
    plugin_instance: str = None,
    index_id: str = None,
    e_index: "EmbeddingIndex" = None,
    reindex: bool = True,
    space_id: str = None,
    space_handle: str = None,
    space: Any = None,
) -> "EmbeddingIndex":
    # TODO: This should really be done all on the app, but for now we'll do it in the client
    # to facilitate demos.

    if index_id is None and e_index is not None:
        index_id = e_index.id

    if index_id is None and e_index is None:
        e_index = self.client.create_index(
            plugin_instance=plugin_instance,
            upsert=True,
            space_id=space_id,
            space_handle=space_handle,
            space=space,
        ).data
    elif e_index is None:
        e_index = EmbeddingIndex(client=self.client, id=index_id)

    # We have an index available to us now. Perform the query.
    blocks = self.refresh().data.blocks

    items = []
    for block in blocks:
        item = EmbeddedItem(value=block.text, externalId=block.id, externalType="block")
        items.append(item)

    insert_task = e_index.insert_many(
        items,
        reindex=reindex,
        space_id=space_id,
        space_handle=space_handle,
        space=space,
    )

    insert_task.wait()
    return e_index


File.index = index  # TODO (enias): Q: Why are we adding this as an extension and not directly as a class method



from steamship.data.plugin import PluginType

__copyright__ = "Steamship"
__license__ = "MIT"

from tests import APPS_PATH
from tests.client.operations.test_embed import (
    basic_embedding_search,
    basic_embeddings,
    count_embeddings,
)
from tests.client.operations.test_embedding_index import create_index
from tests.utils.client import get_steamship_client
from tests.utils.deployables import deploy_plugin


def test_e2e_embedder():
    client = get_steamship_client()
    embedder_path = APPS_PATH / "plugins" / "taggers" / "plugin_embedder.py"

    with deploy_plugin(client, embedder_path, PluginType.tagger) as (
        plugin,
        version,
        instance,
    ):
        e1 = client.tag("This is a test", plugin_instance=instance.handle)
        e1.wait()
        assert e1.error is None
        assert count_embeddings(e1.data.file) == 1
        assert len(e1.data.file.blocks[0].tags[0].value["embedding"]) > 1

        e2 = client.tag("This is a test", plugin_instance=instance.handle)
        e2.wait()
        assert e2.error is None
        assert count_embeddings(e2.data.file) == 1
        assert len(e2.data.file.blocks[0].tags[0].value["embedding"]) == len(
            e1.data.file.blocks[0].tags[0].value["embedding"]
        )

        e4 = client.tag("This is a test", plugin_instance=instance.handle)
        e4.wait()
        assert e4.error is None
        assert count_embeddings(e4.data.file) == 1

        # Now lets run all the other embedding tests
        basic_embeddings(client, plugin_instance=instance.handle)
        basic_embedding_search(client, plugin_instance=instance.handle)
        create_index(client, plugin_instance=instance.handle)



from abc import ABC
from typing import Any

from steamship.base import Client
from steamship.plugin.inputs.block_and_tag_plugin_input import BlockAndTagPluginInput
from steamship.plugin.outputs.embedded_items_plugin_output import (
    EmbeddedItemsPluginOutput,
)
from steamship.plugin.service import PluginRequest, PluginService


# Note!
# =====
#
# This is the PLUGIN IMPLEMENTOR's View of an Embedder.
#
# If you are using the Steamship Client, you probably want steamship.client.operations.embedder instead
# of this file.
#
class Embedder(PluginService[BlockAndTagPluginInput, EmbeddedItemsPluginOutput], ABC):
    @classmethod
    def subclass_request_from_dict(
        cls, d: Any, client: Client = None
    ) -> PluginRequest[BlockAndTagPluginInput]:
        return BlockAndTagPluginInput.from_dict(d, client=client)



#
# This is the CLIENT-side abstraction for an app.
#
# If you are implementing an app, see: steamship.app.server.App
#

from dataclasses import dataclass
from typing import Any

from steamship.base import Client, Request, Response


@dataclass
class CreateAppRequest(Request):
    id: str = None
    handle: str = None
    upsert: bool = None


@dataclass
class DeleteAppRequest(Request):
    id: str


@dataclass
class ListPrivateAppsRequest(Request):
    pass


@dataclass
class App:
    client: Client = None
    id: str = None
    handle: str = None

    @staticmethod
    def from_dict(d: Any, client: Client = None) -> "App":
        if "app" in d:
            d = d["app"]
        return App(client=client, id=d.get("id", None), handle=d.get("handle", None))

    @staticmethod
    def create(
        client: Client, handle: str = None, upsert: bool = None
    ) -> "Response[App]":
        req = CreateAppRequest(handle=handle, upsert=upsert)
        return client.post("app/create", payload=req, expect=App)

    def delete(self) -> "Response[App]":
        req = DeleteAppRequest(id=self.id)
        return self.client.post("app/delete", payload=req, expect=App)



import base64
import json
from dataclasses import asdict

import requests

from steamship.data import Block, Tag
from steamship.data.plugin_instance import PluginInstance
from steamship.extension.file import File
from steamship.plugin.inputs.export_plugin_input import ExportPluginInput

from .. import APPS_PATH

__copyright__ = "Steamship"
__license__ = "MIT"

from ..utils.client import get_steamship_client
from ..utils.deployables import deploy_plugin
from ..utils.file import upload_file

EXPORTER_HANDLE = "signed-url-exporter"


def test_e2e_corpus_export():
    client = get_steamship_client()
    version_config_template = dict(
        text_column=dict(type="string"),
        tag_columns=dict(type="string"),
        tag_kind=dict(type="string"),
    )  # TODO (enias): Derive this from Config
    instance_config = dict(  # Has to match up
        text_column="Message",
        tag_columns="Category",
        tag_kind="Intent",
    )
    exporter_plugin_r = PluginInstance.create(
        client=client,
        handle=EXPORTER_HANDLE,
        plugin_handle=EXPORTER_HANDLE,
        upsert=True,
    )
    assert exporter_plugin_r.data is not None
    exporter_plugin = exporter_plugin_r.data
    assert exporter_plugin.handle is not None

    _input = ExportPluginInput(handle="default", type="corpus")
    print(asdict(_input))

    csv_blockifier_path = APPS_PATH / "plugins" / "blockifiers" / "csv_blockifier.py"

    # Make a blockifier which will generate our training corpus
    with deploy_plugin(
        client,
        csv_blockifier_path,
        "blockifier",
        version_config_template=version_config_template,
        instance_config=instance_config,
    ) as (plugin, version, instance):
        with upload_file(client, "utterances.csv") as file:
            assert len(file.refresh().data.blocks) == 0
            # Use the plugin we just registered
            file.blockify(plugin_instance=instance.handle).wait()
            assert len(file.refresh().data.blocks) == 5

            # Now export the corpus
            raw_data_r = exporter_plugin.export(_input)
            assert raw_data_r is not None

            # The results of a corpus exporter are MD5 encoded!
            _ = raw_data_r.data


def test_e2e_corpus_export_with_query():
    client = get_steamship_client()
    exporter_plugin_r = PluginInstance.create(
        client=client,
        handle=EXPORTER_HANDLE,
        plugin_handle=EXPORTER_HANDLE,
        upsert=True,
    )
    assert exporter_plugin_r.data is not None
    exporter_plugin = exporter_plugin_r.data
    assert exporter_plugin.handle is not None

    a = File.create(
        client=client,
        blocks=[
            Block.CreateRequest(text="A", tags=[Tag.CreateRequest(name="BlockTag")]),
            Block.CreateRequest(text="B"),
        ],
    ).data
    assert a.id is not None
    b = File.create(
        client=client,
        blocks=[Block.CreateRequest(text="A"), Block.CreateRequest(text="B")],
        tags=[Tag.CreateRequest(name="FileTag")],
    ).data
    assert b.id is not None

    # Now export the corpus
    _input = ExportPluginInput(query='filetag and name "FileTag"', type="file")
    print(asdict(_input))
    raw_data_r = exporter_plugin.export(_input)
    assert raw_data_r is not None

    # The results of a corpus exporter are MD5 encoded!
    raw_data_r.wait()
    raw_data = raw_data_r.data.data
    # decode base64 to get URL at url json property
    decoded_data = json.loads(base64.b64decode(raw_data))
    url = decoded_data["url"]

    # fetch the URL via requests.get
    content = requests.get(url).text

    # Look at lines of jsonl file
    files = [File.from_dict(json.loads(line)) for line in content.splitlines()]
    assert len(files) == 1
    assert len(files[0].tags) == 1

    a.delete()
    b.delete()



from dataclasses import dataclass
from typing import Any, Dict, Optional

from steamship.base import Client
from steamship.data.file import File


@dataclass
class BlockAndTagPluginOutput:
    # TODO (enias): Review
    file: File.CreateRequest = None

    @staticmethod
    def from_dict(
        d: Any = None, client: Client = None
    ) -> "Optional[BlockAndTagPluginOutput]":
        if d is None:
            return None

        return BlockAndTagPluginOutput(
            file=File.CreateRequest.from_dict(d.get("file", None), client=client)
        )

    def to_dict(self) -> Dict:
        if self.file is None:
            return {}
        return {"file": self.file.to_dict()}



from dataclasses import dataclass
from typing import Any, Generic, List, Type, TypeVar, Union

from steamship.base.base import IResponse
from steamship.base.error import SteamshipError
from steamship.base.metadata import metadata_to_str, str_to_metadata
from steamship.base.request import Request

T = TypeVar("T")


@dataclass
class CreateTaskCommentRequest(Request):
    taskId: str
    externalId: str = None
    externalType: str = None
    externalGroup: str = None
    metadata: str = None
    upsert: bool = None


@dataclass
class ListTaskCommentRequest(Request):
    taskId: str = None
    externalId: str = None
    externalType: str = None
    externalGroup: str = None


@dataclass
class DeleteTaskCommentRequest(Request):
    id: str = None


@dataclass
class TaskComment:
    client: Any = None
    id: str = None
    user_id: str = None
    task_id: str = None
    external_id: str = None
    external_type: str = None
    external_group: str = None
    metadata: Any = None
    created_at: str = None

    @staticmethod
    def create(
        client: Any,  # TODO (Enias): Isn't this Steamship client?
        task_id: str = None,
        external_id: str = None,
        external_type: str = None,
        external_group: str = None,
        metadata: Any = None,
        upsert: bool = True,
    ) -> "IResponse[TaskComment]":
        req = CreateTaskCommentRequest(
            taskId=task_id,
            externalId=external_id,
            externalType=external_type,
            externalGroup=external_group,
            metadata=metadata_to_str(metadata),
            upsert=upsert,
        )
        return client.post(
            "task/comment/create",
            req,
            expect=TaskComment,
        )

    @staticmethod
    def list(
        client: Any,
        task_id: str = None,
        external_id: str = None,
        external_type: str = None,
        external_group: str = None,
    ) -> "IResponse[TaskCommentList]":
        req = ListTaskCommentRequest(
            taskId=task_id,
            externalId=external_id,
            externalType=external_type,
            externalGroup=external_group,
        )
        return client.post(
            "task/comment/list",
            req,
            expect=TaskCommentList,
        )

    def delete(self) -> "IResponse[TaskComment]":
        req = DeleteTaskCommentRequest(self.id)
        return self.client.post(
            "task/comment/delete",
            req,
            expect=TaskComment,
        )

    @staticmethod
    def from_dict(d: Any, client: Any = None) -> "TaskComment":
        return TaskComment(
            client=client,
            id=d.get("id"),
            user_id=d.get("userId"),
            task_id=d.get("taskId"),
            external_id=d.get("externalId"),
            external_type=d.get("externalType"),
            external_group=d.get("externalGroup"),
            metadata=str_to_metadata(d.get("metadata")),
            created_at=d.get("createdAt"),
        )


@dataclass
class TaskCommentList:
    comments: List[TaskComment]

    @staticmethod
    def from_dict(d: Any, client: Any = None) -> "TaskCommentList":
        return TaskCommentList(
            comments=[TaskComment.from_dict(dd, client) for dd in d.get("comments", [])]
        )


class TaskState:
    waiting = "waiting"
    running = "running"
    succeeded = "succeeded"
    failed = "failed"


class TaskType:
    internalApi = "internalApi"
    train = "train"
    infer = "infer"


@dataclass
class TaskRunRequest(Request):
    taskId: str


@dataclass
class TaskStatusRequest(Request):
    taskId: str


@dataclass
class Task(Generic[T]):
    """Encapsulates a unit of asynchronously performed work."""

    client: Any = None  # Steamship client

    task_id: str = None  # The id of this task
    user_id: str = None  # The user who requested this task
    space_id: str = None  # The space in which this task is executing

    input: str = None  # The input provided to the task
    state: str = None  # A value in class TaskState

    status_message: str = None  # User-facing message concerning task status
    status_suggestion: str = None  # User-facing suggestion concerning error remediation
    status_code: str = None  # User-facing error code for support assistance
    status_created_on: str = None  # When the status fields were last set

    task_type: str = None  # A value in class TaskType; for internal routing
    task_executor: str = None  #
    task_created_on: str = None  # When the task object was created
    task_last_modified_on: str = None  # When the task object was last modified

    assigned_worker: str = None  # The worker assigned to complete this task
    started_at: str = None  # When the work on this task began

    max_retries: int = None  # The maximum number of retries allowed for this task
    retries: int = None  # The number of retries already used.

    # This is typed wrong, but I'm not sure how to type it..
    # This is a local object, for use in Python only, that helps deserialize the task
    # to a python object upon completion.
    eventual_result_type: Type[Any] = None

    @staticmethod
    def failure(message: str, suggestion: str, _: str, code: str) -> "Task":
        return Task(
            state=TaskState.failed,
            status_message=message,
            status_suggestion=suggestion,
            status_code=code,
        )

    @staticmethod
    def from_error(error: Union[SteamshipError, Exception]) -> "Task":
        if type(error) == SteamshipError:
            return Task(
                state=TaskState.failed,
                status_message=error.message,
                status_suggestion=error.suggestion,
                status_code=error.code,
            )
        else:
            return Task(state=TaskState.failed, status_message=str(error))

    @staticmethod
    def from_dict(d: Any, client: Any = None) -> "Task":  # TODO (Enias): Review
        """Last resort if subclass doesn't override: pass through."""
        return Task(
            client=client,
            task_id=d.get("taskId"),
            user_id=d.get("userId"),
            space_id=d.get("spaceId"),
            input=d.get("input"),
            state=d.get("state"),
            status_message=d.get("statusMessage"),
            status_suggestion=d.get("statusSuggestion"),
            status_code=d.get("statusCode"),
            status_created_on=d.get("statusCreatedOn"),
            task_type=d.get("taskType"),
            task_executor=d.get("taskExecutor"),
            task_created_on=d.get("taskCreatedOn"),
            task_last_modified_on=d.get("taskLastModifiedOn"),
            assigned_worker=d.get("assignedWorker"),
            started_at=d.get("startedAt"),
            max_retries=d.get("maxRetries"),
            retries=d.get("retries"),
        )

    def to_dict(self) -> dict:
        return dict(
            taskId=self.task_id,
            userId=self.user_id,
            spaceId=self.space_id,
            input=self.input,
            state=self.state,
            statusMessage=self.status_message,
            statusSuggestion=self.status_suggestion,
            statusCode=self.status_code,
            statusCreatedOn=self.status_created_on,
            taskType=self.task_type,
            taskExecutor=self.task_executor,
            taskCreatedOn=self.task_created_on,
            taskLastModifiedOn=self.task_last_modified_on,
            assignedWorker=self.assigned_worker,
            startedAt=self.started_at,
            maxRetries=self.max_retries,
            retries=self.retries,
        )

    def update(self, other: "Task"):
        """Incorporates a `Task` into this object."""
        # TODO (Enias): Simplify with operations on __dict__
        if other is not None:
            self.task_id = other.task_id
            self.user_id = other.user_id
            self.space_id = other.space_id
            self.input = other.input
            self.state = other.state
            self.status_message = other.status_message
            self.status_suggestion = other.status_suggestion
            self.status_code = other.status_code
            self.status_created_on = other.status_created_on
            self.task_type = other.task_type
            self.task_executor = other.task_executor
            self.task_created_on = other.task_created_on
            self.task_last_modified_on = other.task_last_modified_on
            self.assigned_worker = other.assigned_worker
            self.started_at = other.started_at
            self.max_retries = other.max_retries
            self.retries = other.retries
        else:
            self.task_id = None  # TODO (enias): Review typing
            self.user_id = None
            self.space_id = None
            self.input = None
            self.state = None
            self.status_message = None
            self.status_suggestion = None
            self.status_code = None
            self.status_created_on = None
            self.task_type = None
            self.task_executor = None
            self.task_created_on = None
            self.task_last_modified_on = None
            self.assigned_worker = None
            self.started_at = None
            self.max_retries = None
            self.retries = None

    def add_comment(
        self,
        external_id: str = None,
        external_type: str = None,
        external_group: str = None,
        metadata: Any = None,
        upsert: bool = True,
    ) -> IResponse[TaskComment]:
        return TaskComment.create(
            client=self.client,
            task_id=self.task_id,
            external_id=external_id,
            external_type=external_type,
            external_group=external_group,
            metadata=metadata,
            upsert=upsert,
        )

    def list_comments(self) -> IResponse[TaskCommentList]:
        return TaskComment.list(client=self.client, task_id=self.task_id)

    @staticmethod
    def delete_comment(comment: TaskComment = None) -> IResponse[TaskComment]:
        return comment.delete()



import json
import os
from pathlib import Path
from typing import Any, Dict

_configFile = ".steamship.json"

MAX_DEPTH = 40


class Configuration:
    # TODO (enias): Review
    api_key: str = None
    api_base: str = None
    app_base: str = None
    web_base: str = None
    space_id: str = None
    space_handle: str = None
    profile: str = None

    @staticmethod
    def from_dict(d: dict) -> "Configuration":
        if d is None:
            return Configuration()

        return Configuration(
            api_key=d.get("apiKey"),
            api_base=d.get("apiBase"),
            app_base=d.get("appBase"),
            web_base=d.get("webBase"),
            space_id=d.get("spaceId"),
            space_handle=d.get("spaceHandle"),
        )

    def __init__(
        self,
        api_key: str = None,
        api_base: str = None,
        app_base: str = None,
        web_base: str = None,
        space_id: str = None,
        space_handle: str = None,
        profile: str = None,
        config_file: str = None,
        config_dict: dict = None,
    ):
        # First set the profile
        if "STEAMSHIP_PROFILE" in os.environ:
            self.profile = os.getenv("STEAMSHIP_PROFILE")
        if profile is not None:
            self.profile = profile

        # Then load from a file if provided
        if config_file is not None:
            self.load_from_file(config_file, self.profile)
        else:
            self.try_autofinding_files(self.profile)

        self.apply_env_var_overrides()

        if config_dict is not None:
            self.merge_dict(config_dict)

        self.apply_invocation_overrides(
            api_key=api_key,
            api_base=api_base,
            app_base=app_base,
            web_base=web_base,
            space_id=space_id,
            space_handle=space_handle,
        )

        if self.api_base is None:
            self.api_base = "https://api.steamship.com/api/v1/"
        if self.app_base is None:
            self.app_base = "https://steamship.run/"
        if self.web_base is None:
            self.web_base = "https://app.steamship.com/"

        if self.api_base[len(self.api_base) - 1] != "/":
            self.api_base = f"{self.api_base}/"

        if self.app_base[len(self.app_base) - 1] != "/":
            self.app_base = f"{self.app_base}/"

        if self.web_base[len(self.web_base) - 1] != "/":
            self.web_base = f"{self.web_base}/"

    def merge_dict(self, d: Dict[str, Any]):
        api_key = d.get("apiKey")
        if api_key is not None:
            self.api_key = api_key

        api_base = d.get("apiBase")
        if api_base is not None:
            self.api_base = api_base

        app_base = d.get("appBase")
        if app_base is not None:
            self.app_base = app_base

        web_base = d.get("webBase")
        if web_base is not None:
            self.web_base = web_base

        profile = d.get("profile")
        if profile is not None:
            self.profile = profile

        space_id = d.get("spaceId")
        if space_id is not None:
            self.space_id = space_id

        space_handle = d.get("spaceHandle")
        if space_handle is not None:
            self.space_handle = space_handle

    def load_from_file(self, filepath: str, profile: str = None, throw_on_error=True):
        if not os.path.exists(filepath):
            if throw_on_error:
                raise Exception(
                    f"Tried to load configuration file at {filepath} but it did not exist."
                )
            else:
                return

        try:
            with open(filepath, "r") as f:
                s = f.read()
                j = json.loads(s)

                if profile is None:
                    self.merge_dict(j)
                else:
                    if "profiles" not in j or profile not in j["profiles"]:
                        raise Exception(
                            f"Profile {profile} requested but not found in {filepath}"
                        )
                    self.merge_dict(j["profiles"][profile])

        except Exception as err:
            if throw_on_error:
                raise err
            return

    def try_autofinding_files(self, profile: str = None):
        """
        Tries folders from cwd up to root.
        """
        paths = []
        cwd = Path(os.getcwd()).absolute()
        i = 0
        while len(str(cwd)) > 0 and str(cwd) != os.path.sep:
            paths.append(os.path.join(cwd, _configFile))
            cwd = cwd.parent.absolute()
            i += 1
            if i > MAX_DEPTH:
                print("ERROR: Max depth exceeded in config search recursion.")
                break

        paths.append(os.path.join(str(Path.home()), _configFile))
        for filepath in paths:
            if os.path.exists(filepath):
                self.load_from_file(filepath, profile=profile)
                break  # Once we've found it; we're done.

    def apply_env_var_overrides(self):
        """Overrides with env vars"""
        if "STEAMSHIP_API_KEY" in os.environ:
            self.api_key = os.getenv("STEAMSHIP_API_KEY")
        if "STEAMSHIP_API_BASE" in os.environ:
            self.api_base = os.getenv("STEAMSHIP_API_BASE")
        if "STEAMSHIP_APP_BASE" in os.environ:
            self.app_base = os.getenv("STEAMSHIP_APP_BASE")
        if "STEAMSHIP_WEB_BASE" in os.environ:
            self.web_base = os.getenv("STEAMSHIP_WEB_BASE")
        if "STEAMSHIP_SPACE_ID" in os.environ:
            self.space_id = os.getenv("STEAMSHIP_SPACE_ID")
        if "STEAMSHIP_SPACE_HANDLE" in os.environ:
            self.space_handle = os.getenv("STEAMSHIP_SPACE_HANDLE")

    def apply_invocation_overrides(
        self,
        api_key: str = None,
        api_base: str = None,
        app_base: str = None,
        web_base: str = None,
        space_id: str = None,
        space_handle: str = None,
    ):
        if api_key is not None:
            self.api_key = api_key
        if api_base is not None:
            self.api_base = api_base
        if app_base is not None:
            self.app_base = app_base
        if web_base is not None:
            self.web_base = web_base
        if space_id is not None:
            self.space_id = space_id
        if space_handle is not None:
            self.space_handle = space_handle



from steamship.data.plugin import TrainingPlatform
from steamship.data.plugin_instance import PluginInstance
from steamship.plugin.inputs.export_plugin_input import ExportPluginInput
from steamship.plugin.inputs.training_parameter_plugin_input import (
    TrainingParameterPluginInput,
)

from .. import APPS_PATH

__copyright__ = "Steamship"
__license__ = "MIT"

from ..utils.client import get_steamship_client
from ..utils.deployables import deploy_plugin
from ..utils.file import upload_file

EXPORTER_HANDLE = "signed-url-exporter"


def test_e2e_trainable_tagger_ecs_training():
    client = get_steamship_client()

    version_config_template = dict(
        text_column=dict(type="string"),
        tag_columns=dict(type="string"),
        tag_kind=dict(type="string"),
    )
    instance_config = dict(
        text_column="Message", tag_columns="Category", tag_kind="Intent"
    )

    exporter_plugin_r = PluginInstance.create(
        client=client,
        handle=EXPORTER_HANDLE,
        plugin_handle=EXPORTER_HANDLE,
        upsert=True,
    )
    assert exporter_plugin_r.data is not None
    exporter_plugin = exporter_plugin_r.data
    assert exporter_plugin.handle is not None

    csv_blockifier_path = APPS_PATH / "plugins" / "blockifiers" / "csv_blockifier.py"
    trainable_tagger_path = (
        APPS_PATH / "plugins" / "taggers" / "plugin_trainable_tagger.py"
    )

    # Make a blockifier which will generate our training corpus
    with deploy_plugin(
        client,
        csv_blockifier_path,
        "blockifier",
        version_config_template=version_config_template,
        instance_config=instance_config,
    ) as (plugin, version, instance):
        with upload_file(client, "utterances.csv") as file:
            assert len(file.refresh().data.blocks) == 0
            # Use the plugin we just registered
            file.blockify(plugin_instance=instance.handle).wait()
            assert len(file.refresh().data.blocks) == 5

            # Now make a trainable tagger to train on those tags
            with deploy_plugin(
                client,
                trainable_tagger_path,
                "tagger",
                training_platform=TrainingPlatform.managed,
            ) as (tagger, taggerVersion, taggerInstance):
                # Now train the plugin
                training_request = TrainingParameterPluginInput(
                    pluginInstance=taggerInstance.handle,
                    exportRequest=ExportPluginInput(
                        pluginInstance=EXPORTER_HANDLE, type="corpus", handle="default"
                    ),
                )

                train_result = taggerInstance.train(training_request)
                train_result.wait()



import io
from dataclasses import dataclass
from typing import Any, Optional, Union

from steamship.base import Client, MimeTypes
from steamship.base.binary_utils import flexi_create


@dataclass
class RawDataPluginOutput:
    """Represents mime-typed raw data that can be returned to the engine.

    As a few examples, you can return:
    - Raw text: RawDataPluginOutput(string=raw_text, MimeTypes.TXT)
    - Markdown text: RawDataPluginOutput(string=markdown_text, MimeTypes.MKD)
    - A PNG image: RawDataPluginOutput(bytes=png_bytes, MimeTypes.PNG)
    - A JSON-serializable Dataclass: RawDataPluginOutput(json=dataclass, MimeTypes.JSON)
    - Steamship Blocks: RawDataPluginOutput(json=file, MimeTypes.STEAMSHIP_BLOCK_JSON)

    The `data` field of this object will ALWAYS be Base64 encoded by the constructor. This ensures that the object
    is always trivially JSON-serializable over the wire, no matter what it contains.

    The `mimeType` field of this object should always be filled in if known. The Steamship Engine makes use of it
    to proactively select defaults for handling the data returned.
    """
    data: Optional[str] = None        # Note: This is **always** Base64 encoded.
    mimeType: Optional[str] = None

    def __init__(
            self,
            base64string: str = None,
            string: str = None,
            bytes: Union[bytes, io.BytesIO] = None,
            json: Any = None,
            mime_type: str = None
    ):
        if base64string is not None:
            self.data = base64string
            self.mime_type = mime_type or MimeTypes.BINARY
        else:
            # Base64-encode the data field.
            self.data, self.mimeType, encoding = flexi_create(
                base64string=base64string,
                string=string,
                json=json,
                bytes=bytes,
                mime_type=mime_type,
                force_base64=True
            )

    @staticmethod
    def from_dict(d: any, client: Client = None) -> "RawDataPluginOutput":
        # We expect the serialized version of this object to always include a Base64 encoded string,
        # so we present it to the constructor as such.
        return RawDataPluginOutput(
            base64string=d.get('data', None),
            mime_type=d.get('mimeType', None)
        )

    def to_dict(self) -> dict:
        return dict(data=self.data, mimeType=self.mimeType)
